package io.gitlab.jfronny.commons.http.client;

import io.gitlab.jfronny.commons.serialize.Serializer;

import java.io.*;
import java.lang.reflect.Type;
import java.net.*;
import java.net.http.HttpRequest;
import java.net.http.HttpResponse;
import java.net.http.HttpTimeoutException;
import java.nio.charset.StandardCharsets;
import java.util.*;
import java.util.function.Predicate;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class HttpClient {
    private static String userAgent = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36";
    private static final String PROXY_AUTH;
    private static final java.net.http.HttpClient CLIENT;

    static {
        // Enables HTTPS proxying
        System.setProperty("jdk.http.auth.tunneling.disabledSchemes", "");

        java.net.http.HttpClient.Builder clientBuilder = java.net.http.HttpClient.newBuilder().followRedirects(java.net.http.HttpClient.Redirect.ALWAYS);

        String host = System.getProperty("http.proxyHost");
        String port = System.getProperty("http.proxyPort");
        if (host != null && port != null) {
            if (port.matches("\\d+")) {
                clientBuilder.proxy(ProxySelector.of(new InetSocketAddress(host, Integer.parseInt(port))));
            } else {
                System.err.println("Could not parse proxy settings: Port is not a number");
            }
        }

        String name = System.getProperty("http.proxyUserName");
        String pass = System.getProperty("http.proxyUserPassword");
        if (name != null && pass != null) {
            PROXY_AUTH = "Basic " + new String(Base64.getEncoder().encode((name + ":" + pass).getBytes()));
        } else {
            PROXY_AUTH = null;
        }

        CLIENT = clientBuilder.build();
    }

    public static void setUserAgent(String hostname) {
        if (hostname == null || hostname.isEmpty()) throw new IllegalArgumentException("Hostname cannot be empty");
        userAgent = hostname;
    }

    private enum Method {
        GET,
        POST,
        PUT,
        PATCH,
        DELETE
    }

    public static class Request {
        private static final Predicate<String> CURSEFORGE_API = Pattern.compile("(?:http(s)?://)?addons-ecs\\.forgesvc\\.net/api/+").asMatchPredicate();
        private final String url;
        private final HttpRequest.Builder builder;
        private Method method;
        private int sent = 0;
        private int retryAfterDefault = 5000;
        private int retryAfterMax = 15000;
        private int retryLimit = 3;
        private List<Exception> retryExceptions = null;

        public Request(Method method, String url) throws URISyntaxException {
            this.url = url.replace(" ", "%20");
            this.builder = HttpRequest.newBuilder()
                    .uri(new URI(this.url));
            this.method = method;
            userAgent(userAgent);
        }

        public Request bearer(String token) {
            builder.header("Authorization", "Bearer " + token);

            return this;
        }

        public Request header(String name, String value) {
            builder.header(name, value);
            return this;
        }

        public Request setHeader(String name, String value) {
            builder.setHeader(name, value);
            return this;
        }

        public Request configureRetryAfter(int defaultDelay, int maxDelay) {
            if (defaultDelay < 1) throw new IllegalArgumentException("defaultDelay must be greater than zero");
            if (maxDelay < defaultDelay) throw new IllegalArgumentException("maxDelay must be greater than or equal to defaultDelay");
            retryAfterDefault = defaultDelay;
            retryAfterMax = maxDelay;
            return this;
        }

        public Request setRetryLimit(int limit) {
            if (limit < 0) throw new IllegalArgumentException("limit must be greater than or zero");
            retryLimit = limit;
            return this;
        }

        public Request userAgent(String value) {
            return setHeader("User-Agent", value);
        }

        public Request bodyString(String string) {
            builder.header("Content-Type", "text/plain");
            builder.method(method.name(), HttpRequest.BodyPublishers.ofString(string));
            method = null;

            return this;
        }

        public Request bodyForm(String string) {
            builder.header("Content-Type", "application/x-www-form-urlencoded");
            builder.method(method.name(), HttpRequest.BodyPublishers.ofString(string));
            method = null;

            return this;
        }

        public Request bodyForm(Map<String, String> entries) {
            return bodyForm(entries.entrySet()
                    .stream()
                    .map(entry -> URLEncoder.encode(entry.getKey(), StandardCharsets.UTF_8) + '=' + URLEncoder.encode(entry.getValue(), StandardCharsets.UTF_8))
                    .collect(Collectors.joining("&")));
        }

        public Request bodyJson(String string) {
            builder.header("Content-Type", "application/json");
            builder.method(method.name(), HttpRequest.BodyPublishers.ofString(string));
            method = null;

            return this;
        }

        public Request bodySerialized(Object object) throws IOException {
            Serializer serializer = Serializer.getInstance();
            builder.header("Content-Type", serializer.getFormatMime());
            builder.method(method.name(), HttpRequest.BodyPublishers.ofString(serializer.serialize(object)));
            method = null;

            return this;
        }

        private <T> T _send(String accept, HttpResponse.BodyHandler<T> responseBodyHandler) throws IOException {
            sent++;
            if (sent > retryLimit) {
                IOException e = new IOException("Attempted to reconnect/redirect " + sent + " times, which is more than the permitted " + retryLimit + ". Stopping");
                if (retryExceptions != null) for (Exception ex : retryExceptions) e.addSuppressed(ex);
                throw e;
            }
            builder.header("Accept", accept);
            if (method != null) builder.method(method.name(), HttpRequest.BodyPublishers.noBody());
            if (PROXY_AUTH != null) builder.header("Proxy-Authorization", PROXY_AUTH);

            HttpResponse<T> res;
            try {
                res = CLIENT.send(builder.build(), responseBodyHandler);
            } catch (InterruptedException e) {
                throw new IOException("Could not send request", e);
            } catch (IOException e) {
                String message = e.getMessage();
                if (message != null && message.contains("GOAWAY received")) {
                    return handleRetryAfter(accept, responseBodyHandler, retryAfterDefault);
                } else throw new IOException("Could not send request", e);
            }
            if (res.statusCode() / 100 == 2) return res.body();
            Optional<String> location = res.headers().firstValue("location");
            Optional<Integer> retryAfter = res.headers().firstValue("Retry-After").flatMap(s -> {
                try {
                    return Optional.of(Integer.parseInt(s));
                } catch (NumberFormatException e) {
                    return Optional.empty();
                }
            });
            final String exceptionSuffix = " (URL=" + url + ")";
            return switch (res.statusCode()) {
                case 429 -> {
                    // Rate limit
                    yield handleRetryAfter(accept, responseBodyHandler, retryAfter.map(s -> s * 1000).orElse(retryAfterDefault));
                }
                case 302, 307 -> {
                    // Redirect
                    if (location.isPresent() && method == Method.GET) {
                        try {
                            yield get(location.get())._send(accept, responseBodyHandler);
                        } catch (URISyntaxException e) {
                            throw new IOException("Could not follow redirect" + exceptionSuffix, e);
                        }
                    }
                    throw new IOException("Unexpected redirect: " + res.statusCode() + exceptionSuffix);
                }
                case 500, 502, 503, 504, 507 -> {
                    // CurseForge serverside error
                    if (CURSEFORGE_API.test(url)) {
                        yield handleRetryAfter(accept, responseBodyHandler, Math.min(1000, retryAfterMax));
                    }
                    throw new IOException("Unexpected serverside error: " + res.statusCode() + exceptionSuffix);
                }
                case 404 -> throw new FileNotFoundException("Didn't find anything under that url" + exceptionSuffix);
                default -> throw new IOException("Unexpected return method: " + res.statusCode() + exceptionSuffix);
            };
        }

        private <T> T handleRetryAfter(String accept, HttpResponse.BodyHandler<T> responseBodyHandler, int millis) throws IOException {
            if (millis > retryAfterMax) throw new HttpTimeoutException("Wait time specified by Retry-After is too long: " + millis);
            try {
                Thread.sleep(millis);
            } catch (InterruptedException e) {
                throw new IOException("Could not sleep before resending request" + e);
            }
            return this._send(accept, responseBodyHandler);
        }

        public void send() throws IOException {
            _send("*/*", HttpResponse.BodyHandlers.discarding());
        }

        public InputStream sendInputStream() throws IOException {
            return _send("*/*", HttpResponse.BodyHandlers.ofInputStream());
        }

        public Reader sendReader() throws IOException {
            return new InputStreamReader(sendInputStream());
        }

        public String sendString() throws IOException {
            return _send("*/*", HttpResponse.BodyHandlers.ofString());
        }

        public Stream<String> sendLines() throws IOException {
            return _send("*/*", HttpResponse.BodyHandlers.ofLines());
        }

        public <T> T sendSerialized(Type type) throws IOException {
            Serializer serializer = Serializer.getInstance();
            InputStream in = _send(serializer.getFormatMime(), HttpResponse.BodyHandlers.ofInputStream());
            try {
                return in == null ? null : serializer.deserialize(new InputStreamReader(in), type);
            } catch (Serializer.SerializeException e) {
                throw new IOException("Could not deserialize", e);
            }
        }

        private String getString(Object a) throws IOException {
            if (a instanceof InputStream s) return new String(s.readAllBytes());
            if (a instanceof String s) return s;
            if (a instanceof Stream s) return ((Stream<String>) s).collect(Collectors.joining());
            return "";
        }
    }

    public static Request get(String url) throws URISyntaxException {
        return new Request(Method.GET, url);
    }

    public static Request post(String url) throws URISyntaxException {
        return new Request(Method.POST, url);
    }

    public static Request put(String url) throws URISyntaxException {
        return new Request(Method.PUT, url);
    }

    public static Request patch(String url) throws URISyntaxException {
        return new Request(Method.PATCH, url);
    }

    public static Request delete(String url) throws URISyntaxException {
        return new Request(Method.DELETE, url);
    }
}
